extends CharacterBody2D

# Animation variables, in the player code these are set, then the animation-tree
# variables are updated, to make things more readable and have everything
# related to the animation tree in one place.
var just_wavedashed := false:
	get: return _reset_on_get("just_wavedashed")
var just_jumped := false:
	get: return _reset_on_get("just_jumped")
var just_walljumped := false:
	get: return _reset_on_get("just_walljumped")
var just_doublejumped := false:
	get: return _reset_on_get("just_doublejumped")
var just_dashed := false:
	get: return _reset_on_get("just_dashed")
var just_took_damage := false:
	get: return _reset_on_get("just_took_damage")
var just_died := false:
	get: return _reset_on_get("just_died")
var just_attacked := false:
	get: return _reset_on_get("just_attacked")
var just_respawned := false:
	get: return _reset_on_get("just_respawned")

var standing_still := true
var falling := false
var on_floor := true
var on_wall := true

func _update_anim_vars():
	standing_still = is_zero_approx(velocity.x)
	falling = is_zero_approx(velocity.y)
	on_floor = is_on_floor()
	on_wall = is_on_wall()

## Gets set as false when read as true.
func _reset_on_get(field):
	var old = self[field]
	if self[field]:
		self[field] = false
	return old

func current_state():
	return stm.state

@onready var stm = $StateMachine
@onready var animation_player = $AnimationPlayer
@onready var animation_tree = $AnimationTree
@onready var player_character_sprite = $PlayerCharacterSprite
@onready var camera = $Detatch/Camera
@onready var follow_camera = $FollowCamera
@onready var air = $StateMachine/Air
@onready var run = $StateMachine/Run
@onready var dash = $StateMachine/Dash
@onready var wallslide = $StateMachine/Wallslide
@onready var death = $StateMachine/Death
@onready var shape = $CollisionShape
@onready var fade_to_black = $UI/Root/FadeToBlack
@onready var health_ui = $UI/Root/Health
@onready var hurt_box = $HurtBox

@onready var hitbox_up = $Hitboxes/Up
@onready var hitbox_down = $Hitboxes/Down
@onready var hitbox_side = $Hitboxes/Side

# Powerup enable/disable
@export var double_jump_enabled := true
@export var dash_enabled := true
@export var wavedash_enabled := true
@export var wallslide_enabled := true
@export var dashjump_enabled := true

## The time frame in which a buffered jump may occur
@export
var buffer_jump := 0.0
var buffer_jump_tmp := 0.0

## The time frame in which the player is still in the run state even though
## they are not on the ground anymore, and when a coyote jump may happen
@export
var coyote_jump := 0.0
var coyote_jump_tmp := coyote_jump

## The delay between pressing jump in the air and executing a double jump, mainly
## used so that buffered jumps can happend when having a double jump
@export
var double_jump_delay := 0.0
var double_jump_delay_tmp := 0.0
var double_jump_delay_initiated := false

@export var dash_time: float
var dash_time_tmp := dash_time

## The time hitboxes are active when attacking, TODO: make animations control this
@export 
var hit_time: float
var hit_time_tmp: float
var hit_active := false

## The set velocity for pogo jumps
@export
var pogo_jump: float
## The set x velocity for wavedashes
@export
var wavedash_x_velocity: float
## Dashjump can be done at end of dash, this is the timeframe at the end of the dash
@export
var dashjump_timeframe: float

@export var fade_time := 0.3
var current_fade_time := fade_time
var fading := false

@export var health := 5
var current_health := health

var double_jump_used := false
var dash_used := false
# Not an animation variable
var looking_right := true
var in_control := true
# The direction to walk, can be controlled by external actors if !in_control
var input_dir := Vector2.ZERO
var exit_point  ## The point to enter a scene

func _ready():
	stm.state_debug_stat = { "key": "player state" }
	stm.state_debug_log = { "description": "player state" }
	air.jump_debug_stat = { "key": "player jump curve" }
	air.air_run_debug_stat = { "key": "player run curve" }
	run.run_debug_stat = { "key": "player run curve" }
	fade(true, 0)
	if !enter_scene():
		fade(false)

func update_looking_right():
	var dir = get_dir()
	if dir.x < 0:
		looking_right = false
	if dir.x > 0:
		looking_right = true
	follow_camera.looking_right = looking_right

# General structue: mimics querries of state, than acts apon that
func _physics_process(delta):
	_update_anim_vars()
	# Looking right (both animation and logic)
	if !current_state() in [death]:
		update_looking_right()
	# Flip player nodes that want to be flipped
	for child in self.find_children("*"):
		if child.is_in_group("flip_me"):
			child.scale.x = 1 if looking_right else -1
	# Double Jump
	if current_state() in [air] && double_jump_enabled:
		if double_jump_delay_initiated:
			if double_jump_delay_tmp <= 0 && !double_jump_used:
				double_jump_delay_initiated = false
				double_jump_used = true
				just_doublejumped = true
				do_jump()
				Debug.stat_set({ "key": "player did double jump", "indicate": true })
			else:
				double_jump_delay_tmp -= delta
		elif get_jump_pressed():
			double_jump_delay_initiated = true
			double_jump_delay_tmp = double_jump_delay
	else:
		double_jump_delay_initiated = false
	# Start falling
	if current_state() in [run] && not is_on_floor():
		if coyote_jump_tmp <= 0:
			# Fall
			coyote_jump_tmp = coyote_jump
			stm.transition(air)
		else:
			coyote_jump_tmp -= delta
			if get_jump_pressed():
				do_jump()
				Debug.stat_set({ "key": "player did coyote jump", "indicate": true })
				return
	else:
		coyote_jump_tmp = coyote_jump
	# Land
	if current_state() in [air, wallslide] && is_on_floor():
		land()
	# Wallslide
	if current_state() in [air, dash] && is_on_slidable_wall() && !is_on_floor() && wallslide_enabled:
		if (looking_right && get_dir().x > 0) || (!looking_right && get_dir().x < 0):
			if buffer_jump_tmp > 0.0:
				stm.transition(air, { "jump_stats" : current_state().walljump() })
				Debug.stat_set({ "key": "player did buffered walljump", "indicate": true })
			else:
				stm.transition(wallslide)
			buffer_jump_tmp = -1.0
			double_jump_used = false
			dash_used = false
			double_jump_delay_initiated = false
	# Stop wallslide
	if current_state() in [wallslide] && !is_on_slidable_wall():
		stm.transition(air)
	# Wall Jump
	if current_state() in [wallslide] && get_jump_pressed():
		just_walljumped = true
		stm.transition(air, { "jump_stats" : current_state().walljump() })
		return
	# Jump
	if current_state() in [run] && get_jump_pressed():
		do_jump()
		return
	else:
		# Jump buffering
		if get_jump_pressed():
			buffer_jump_tmp = buffer_jump
		else:
			buffer_jump_tmp -= delta
	# Continue jumping
	if current_state() in [air]:
		current_state().continue_jump(Input.is_action_pressed("jump"))
	# Walk left/right
	if current_state() in [run, air]:
		current_state().direction = get_dir().x
	# Do dash
	if current_state() in [run, air] && Input.is_action_just_pressed("dash") && !dash_used \
			&& dash_enabled:
		dash_used = true
		dash_time_tmp = dash_time
		var dir := get_dir()
		if dir == Vector2.ZERO:
			if looking_right: dir.x = 1
			else:             dir.x = -1
		just_dashed = true
		stm.transition(dash)
		current_state().direction = dir
	# Done with dash
	if current_state() in [dash]:
		if dash_time_tmp <= 0:
			if is_on_floor():
				land()
			else:
				stm.transition(air)
		else:
			dash_time_tmp -= delta
			current_state().progression = (dash_time - dash_time_tmp) / dash_time
	# Wavedash
	if current_state() in [dash] && wavedash_enabled:
		if is_on_floor() && get_jump_pressed() && \
			get_dir().y > 0 && get_dir().x != 0:
				just_wavedashed = true
				stm.transition(air, { "jump_stats" : current_state().wavedash_jump() })
				velocity.x = wavedash_x_velocity * get_dir().x
				dash_used = false
				Debug.stat_set({ "key": "player wavedash", "indicate": true })
	# Dashjump
	if current_state() in [dash] && dashjump_enabled:
		if !is_on_floor() && !is_on_wall() && get_jump_pressed() \
			&& dash_time_tmp < dashjump_timeframe:
				stm.transition(air, { "jump_stats" : current_state().dashjump_jump() })
				Debug.stat_set({ "key": "player dashjump", "indicate": true })
	# Environment interaction (spiky ground etc)
	#if !current_state() in [death]:
		#var collision = move_and_collide(dest_vec, true)
	# Interact
	if current_state() in [run] &&\
		get_interact_pressed() &&\
		get_dir().x == 0 &&\
		current_interactable != null:
		current_interactable.interact(self)
	# Hitboxes
	if current_state() in [run, air]:
		if hit_active:
			if hit_time_tmp <= 0:
				hit_active = false
				set_active_hitboxes(false)
			else:
				hit_time_tmp -= delta
		else:
			if looking_right: hitbox_side.get_parent().scale.x = 1
			else:             hitbox_side.get_parent().scale.x = -1
			if Input.is_action_just_pressed("attack"):
				just_attacked = true
				hit_active = true
				hit_time_tmp = hit_time
				var dir = get_dir()
				if dir.y < 0:
					hitbox_up.active = true
				elif dir.y > 0:
					hitbox_down.active = true
				else:
					hitbox_side.active = true
	else:
		set_active_hitboxes(false)
	
	update_health()
	
	current_state().update(delta)
	Debug.stat_set({ "key": "player velocity", "value": str(velocity) })
	
	# Debug stuff
	if Input.is_action_just_pressed("debug_set_respawn"):
		Debug.settings["player_respawn"][get_tree().current_scene.name] = {
			"x" = global_position.x,
			"y" = global_position.y,
		}
		Debug.save_settings()
	if Input.is_action_just_pressed("debug_respawn"):
		if Debug.settings["player_respawn"].has(get_tree().current_scene.name):
			var loaded_pos = Debug.settings["player_respawn"][get_tree().current_scene.name]
			global_position.x = loaded_pos["x"]
			global_position.y = loaded_pos["y"]

func _process(delta: float) -> void:
	fade_to_black.color.a += (1 if fading else -1) * delta / current_fade_time
	fade_to_black.color.a = clamp(fade_to_black.color.a, 0, 1)
	fade_to_black.visible = fade_to_black.color.a > 0.0

func is_on_slidable_wall():
	var dest_vec := Vector2.ZERO
	if looking_right:
		dest_vec.x += 1
	else:
		dest_vec.x -= 1
	var collision = move_and_collide(dest_vec, true)
	if collision != null:
		if "no_wallslide" in collision.get_collider():
			return !collision.get_collider().no_wallslide
		else:
			return true
	else:
		return false

# Helper function
func land():
	double_jump_used = false
	dash_used = false
	if buffer_jump_tmp > 0:
		buffer_jump_tmp = 0
		double_jump_delay_initiated = false
		just_jumped = true
		stm.transition(air, { "jump_stats" : current_state().jump() })
		Debug.stat_set({ "key": "player jumped buffered jump", "indicate": true })
	else:
		stm.transition(run)

## Disables/enables hitboxes
func set_active_hitboxes(active: bool):
	for hitbox in [hitbox_up, hitbox_down, hitbox_side]:
		hitbox.active = active

## Disables/enables hurtbox
func set_active_hurtbox(active: bool):
	hurt_box.active = active

func do_jump():
	just_jumped = true
	stm.transition(air, { "jump_stats" : current_state().jump() })

func get_jump_pressed():
	return Input.is_action_just_pressed("jump")

func get_interact_pressed():
	return Input.is_action_just_pressed("interact")

func get_dir() -> Vector2:
	if !in_control:
		return input_dir
	else:
		var analog := Vector2.ZERO
		if Input.is_action_pressed("walk_left_analog"):
			analog.x += -1
		if Input.is_action_pressed("walk_right_analog"):
			analog.x += 1
		if Input.is_action_pressed("walk_up_analog"):
			analog.y += -1
		if Input.is_action_pressed("walk_down_analog"):
			analog.y += 1
		var digital = Vector2(Input.get_axis("walk_left", "walk_right"), Input.get_axis("walk_up", "walk_down"))
		if analog != Vector2.ZERO: return analog
		else:                      return digital

func apply_camera(c):
	camera.apply_camera(c)

func remove_camera(c):
	camera.remove_camera(c)

func reset_camera():
	camera.reset_camera()

func enable_camera(enable: bool):
	camera.enabled = enable

func snap_to_floor():
	var space_state = get_world_2d().direct_space_state
	var query = PhysicsRayQueryParameters2D.create(global_position, global_position + Vector2(0, 600))
	var result = space_state.intersect_ray(query)
	if result:
		global_position.y = result.position.y - (shape.shape.size.y /2) - shape.position.y
		velocity.y = 0

func fade(fade_in: bool, time: float = fade_time):
	fading = fade_in
	current_fade_time = time

func _on_down_hit(_data):
	dash_used = false
	double_jump_used = false
	velocity.y = - pogo_jump

func update_health():
	health_ui.text = ""
	for _i in current_health:
		health_ui.text += "❤️"

func _on_hurt_box_hurt(data):
	just_took_damage = true
	current_health -= data.damage
	if data.knockback != Vector2.ZERO:
		velocity = data.knockback
	Debug.stat_set({"key": "player health", "value": str(current_health), "indicate": true})
	if current_health <= 0:
		just_died = true
		set_active_hitboxes(false)
		set_active_hurtbox(false)
		stm.transition(death)
		fade(true, fade_time)
		await get_tree().create_timer(fade_time).timeout
		just_respawned = true
		current_health = health
		set_active_hurtbox(true)
		stm.transition(air)
		var found_exit := enter_scene()
		if !found_exit:
			fade(false, fade_time)

## Find a place to enter in the currently loaded scene
## Returns true if an exit point were found
func enter_scene() -> bool:
	for n in get_tree().get_nodes_in_group("scene_transition"):
		if n.name == exit_point:
			n.exit_here(self)
			return true
	# If the player didn't enter through any door
	for n in get_tree().get_nodes_in_group("scene_transition"):
		if n.default_entrance:
			n.exit_here(self)
			return true
	return false

## When entering a scene, an exit point calls this
func exit_here(n):
	n.exit_here(self)

var current_interactable = null
var interact_arrow = preload("res://global/ui/InteractArrow.tscn")
func _on_interact_box_area_entered(area):
	if area.interactable:
		if current_interactable != null:
			current_interactable.get_node("InteractArrow").queue_free()
		current_interactable = area
		var arrow_instance = interact_arrow.instantiate()
		current_interactable.add_child(arrow_instance)

func _on_interact_box_area_exited(area):
	if current_interactable == area:
		current_interactable.get_node("InteractArrow").queue_free()
		current_interactable = null

## The ground sniffer is an area that initiates interactions with
## other collision object. This is needed since interactions
## between two collision objects cannot be detected without doing
## manual collision checks.
func _on_ground_sniffer_body_entered(body: Node2D):
	if body.has_method("actor_enter"):
		body.actor_enter(hurt_box)
func _on_ground_sniffer_body_exited(body: Node2D) -> void:
	if body.has_method("actor_exit"):
		body.actor_exit(hurt_box)
