extends Node2D

## Handles fading the screen, taking control from the player

@export_file("*.tscn") var scene: String
@export var exit_point: String     ## Name of node in the scne transitioning to where "who" exits
@export var walk := true
@export var enter_right := true
@export var enter_time := 0.3
@export var exit_time := 0.3
@export var snap_to_floor := true
## Accessed by other scripts
@export var interactable := false
## Does walking into the area activate the transition?
@export var walkable := true
@export var default_entrance := true

const INIT_GRACE := 0.1
var init_grace_tmp := INIT_GRACE

# An actor interacting with this triggers a scene transtion
func interact(n):
	transition(n)

func transition(n):
	if scene != "" && init_grace_tmp < 0:
		n.exit_point = exit_point
		n.fade(true, enter_time)
		await make_walk(n, enter_right, enter_time)
		get_tree().change_scene_to_file.call_deferred(scene)

func _ready():
	var nodes = get_tree().get_nodes_in_group("cross_scene")
	for n in nodes:
		if n.exit_point == name || name == "SceneTransition":
			# The node in question decides what to do
			n.exit_here(self)

func _physics_process(delta):
	init_grace_tmp -= delta

## Makes an actor exit at this point
func exit_here(n):
	init_grace_tmp = INIT_GRACE
	n.global_position = global_position
	n.velocity = Vector2.ZERO
	n.reset_camera()
	# Make actor enter loose control and walk for walk_time seconds
	if exit_time != 0:
		make_walk(n, !enter_right, exit_time)
	await get_tree().create_timer(0.2).timeout
	n.fade(false, exit_time)
	if snap_to_floor:
		n.snap_to_floor()

func make_walk(n, walk_right, time):
	n.in_control = false
	if walk:
		n.input_dir.x = 1 if walk_right else -1
	else:
		n.input_dir.x = 0
	await get_tree().create_timer(time).timeout
	n.input_dir.x = 0
	n.in_control = true


func _on_body_entered(body):
	if walkable:
		transition(body)
