extends Node

@onready var actor = get_parent().get_parent()

## Curve determining how fast the actor falls when not entering state from a jump
@export
var falling_curve: Curve

@export var air_run_stats: RunStats
@export var double_jump_stats: JumpStats
@export var wall_jump_stats: JumpStats

## The x-velocity set when doing a walljump
@export
var x_launch: float

var current_jump_stats: JumpStats = null

# Set by owner
var released_jump := false
var direction := 0.0
var jump_debug_stat = null
var air_run_debug_stat = null

var jump_time_tmp := 0.0

func enter(args):
	jump_time_tmp = 0.0
	if args.has("jump_stats"):
		released_jump = false
		current_jump_stats = args["jump_stats"]
	else:
		current_jump_stats = null

func update(delta):
	if current_jump_stats != null:
		actor.velocity.y = current_jump_stats.jump(
			actor.velocity.y, jump_time_tmp, released_jump, delta, jump_debug_stat)
	else:
		# When no jump
		actor.velocity.y -= falling_curve.sample(jump_time_tmp) * delta
	# Run in the air
	actor.velocity.x = air_run_stats.run(actor.velocity.x, direction, delta, air_run_debug_stat)
	actor.move_and_slide()
	jump_time_tmp += delta

# Release jump or not, called by owner
func continue_jump(should_continue):
	released_jump = not should_continue

# Double jump
func jump():
	actor.velocity.y = - double_jump_stats.initial_velocity
	return double_jump_stats

func walljump():
	if actor.looking_right: actor.velocity.x -= x_launch
	else:                   actor.velocity.x += x_launch
	actor.velocity.y = - wall_jump_stats.initial_velocity
	return wall_jump_stats
