extends Node2D
## Generic hitbox, lingering or not

signal hit(data)  ## If actor using this wants to know if something is hit

## Use this to activate / deactivate the hitbox
## to make sure that the tracked hurtboxes are reset.
@export var active := true:
	set(v):
		active = v
		process_mode = PROCESS_MODE_INHERIT if active else PROCESS_MODE_DISABLED
		visible = active
		if !active:
			hits = {}

@export var lingering := true ## Does it continue to deal damage?
@export var damage := 1
@export var knockback: Vector2

# During one hit, the same hurtbox can only be hit once
var hits := {}

## Hitting something emits a signal and calls hit on the victim
func _hit_something(what):
	if active:
		var data := {
			"damage": damage,
			"hitbox": self,
			"knockback": Vector2(
				knockback.x if what.global_position.x - global_position.x > 0 
						else -knockback.x, 
				knockback.y
			)
		}
		what.hit(data)
		hit.emit(data)

func _process(_delta):
	if lingering:
		for node in hits:
			_hit_something(node)

## In a typical Area2D, this should be connected to area_entered.
## If a this is a tilemap, the actor has to use an Area2D to detect
## and call this method.
func actor_enter(area):
	if hits.has(area):
		return
	hits[area] = true
	_hit_something(area)

## Same as actor_enter, just as important to call this.
func actor_exit(area):
	hits.erase(area)
