extends Resource
class_name RunStats

## The acceleration (y) at different speeds (x), when walking forward.
## x at where the curve corsses y=0 is the implicit maximum x velocity
@export
var accel_curve: Curve
## Same but applied when walking in the opposite direction of the momentum.
@export
var turn_curve: Curve
## Same but when stopping.
@export
var stop_curve: Curve
## The "width" of the curves, higher values will have longer speedups and slowdowns
@export
var time := 1.0

func _init(p_accel_curve = null, p_turn_curve = null, p_stop_curve = null):
	accel_curve = p_accel_curve
	turn_curve = p_turn_curve
	stop_curve = p_stop_curve

## Utility function to be used by clients.
## Optional stat_key sets a stat with vale of internally used curve.
func run(xvel: float, dir: float, delta: float, stat_key = null) -> float:
	# Stopping speed
	if sign(dir) == 0:
		Debug.maybe_put_stat(stat_key, "stopping")
		return move_toward(xvel, 0, stop_curve.sample(abs(xvel)/time) * delta)
	# Forward speed
	elif sign(dir) == sign(xvel) or sign(xvel) == 0:
		Debug.maybe_put_stat(stat_key, "forward")
		return xvel + sign(dir) * accel_curve.sample(abs(xvel)/time) * delta
	# Turning speed
	else:
		Debug.maybe_put_stat(stat_key, "turning")
		return xvel + sign(dir) * turn_curve.sample(abs(xvel)/time) * delta
